/* Copyright (C) 2005 E.Berta

tarfixer is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

tarfixer is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/


#define _LARGEFILE_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <string.h>
#include <bzlib.h>
#include <zlib.h>
#include <ctype.h>
#include <map>
#include <getopt.h>

using namespace std;

map<int, unsigned int>  usefilelength;

bool on_2nd_pass = false;
bool need_2nd_pass = false;
int  numhits = 0;
int  just_parse = 0;

void formatoctal(char * ostr, uint osize, uint num) {

   uint i = osize - 1;

   ostr[i] = '\0';
   do
   {
      ostr[--i] = '0' + (num & ((1 << 3) - 1));
      num >>= 3;
   } while (i);
 
}

//gzFile gzf;
//BZFILE* bzf;
//FILE* f;
typedef enum {NORMALMODE, BZMODE, ZIPMODE} comp_et;
comp_et w_compmode = NORMALMODE;
comp_et r_compmode = NORMALMODE;

void*  fileopen(char* name, char* mode = "r") 
{
   void* inpfile = NULL;
   comp_et compmode; 
   if(mode[0] == 'r') compmode = r_compmode;
   else if(mode[0] == 'w') {
     compmode = w_compmode;
     if(just_parse) return (void*)1;
   } else {
      printf("Bad mode in fileopen\n");
      abort();
   }

   switch(compmode) {
   case NORMALMODE:
      inpfile = fopen(name, mode);
      break;
   case BZMODE:
      inpfile = BZ2_bzopen(name, mode);
      break;
   case ZIPMODE:
      inpfile = gzopen(name, mode);
      break;      
   }
   return inpfile; 
}

size_t fileread(void* inpfile, char* buf, size_t bytes)
{
   switch(r_compmode) {
   case NORMALMODE:
      return fread(buf, 1, bytes, (FILE*)inpfile);
      break;
   case BZMODE:
      return BZ2_bzread(inpfile, buf, bytes);
      break;
   case ZIPMODE:
      return gzread(inpfile, buf, bytes);
      break;      
   }
}

size_t filewrite(void* outfile, char* buf, size_t bytes)
{

   if(just_parse) return bytes;
   if(need_2nd_pass && !on_2nd_pass) return bytes;

   switch(w_compmode) {
   case NORMALMODE:
      return fwrite(buf, 1, bytes, (FILE*)outfile);
      break;
   case BZMODE:
      return BZ2_bzwrite(outfile, buf, bytes);
      break;
   case ZIPMODE:
      return gzwrite(outfile, buf, bytes);
      break;      
   }
}

int fileclose(void* outfile, char mode)
{
   comp_et compmode;
   if(mode == 'r') compmode = r_compmode;
   else if(mode == 'w') {
     compmode = w_compmode;
     if(just_parse) return 0; 
   } else {
      printf("Bad mode in fileclose\n");
      abort();
   }


   switch(compmode) {
   case NORMALMODE:
      return fclose((FILE*)outfile);
      break;
   case BZMODE:
      BZ2_bzclose(outfile);
      return 0;
      break;
   case ZIPMODE:
      return gzclose(outfile);
      break;      
   }
}


typedef struct posix_header
{				/* byte offset */
   char name[100];		/*   0 */
   char mode[8];			/* 100 */
   char uid[8];			/* 108 */
   char gid[8];			/* 116 */
   char size[12];		/* 124 */
   char mtime[12];		/* 136 */
   char chksum[8];		/* 148 */
   char typeflag;		/* 156 */
   char linkname[100];		/* 157 */
   char magic[6];		/* 257 */
   char version[2];		/* 263 */
   char uname[32];		/* 265 */
   char gname[32];		/* 297 */
   char devmajor[8];		/* 329 */
   char devminor[8];		/* 337 */
   char prefix[155];		/* 345 */
				/* 500 */
} tarheader;

char readbuffer[2048];
char writebuffer[20480];
int wb_in = 0;
int wb_out = 0;

int roffset = 0;

ulong  last_totalread = 0;
size_t readnum = 0;
size_t last_read = 0;
int last_roffset = 0;

char tmagic[] = "ustar";
int  toffset = 256;
int  tat = 0;
int  tmagicsize = 5;

tarheader lastheaders[10];
int atheader = 0;

typedef unsigned long int ulong;
ulong expected_offset = 0;

ulong last_foffset = 0;
void* inpfile;
void* outfile;

void add_wbchar(char inchar) {
  
   writebuffer[wb_in++] = inchar;
   if(wb_in == 20480) wb_in = 0;
   if(wb_in == wb_out) {
      //flush half of it
      wb_out += 10240;
      if(wb_out >= 20480) {
         filewrite(outfile, &(writebuffer[wb_in]), 20480-wb_in);
         wb_out -= 20480;
         filewrite(outfile, writebuffer, wb_out);
      } else {
         filewrite(outfile, &(writebuffer[wb_in]), wb_out-wb_in);
      }
 
   }

}

bool tar_within_tar = false;

void scanchar(char inchar, int cloc) 
{

   add_wbchar(inchar);

   if(inchar == tmagic[tat]) { 
      tat++;
      if(tat == 5) {
         tat = 0;
         int offset = last_roffset + cloc - 5 - toffset;
         int left = sizeof(tarheader); //500;
	 if(offset < 0) offset += 2048;
         int readin = (left < (2048-offset)) ? left : (2048-offset);
    
         memcpy(&(lastheaders[atheader]), &(readbuffer[offset]), readin);
         left -= readin;
         memcpy((char*)(&(lastheaders[atheader])) + readin,
                &(readbuffer[0]), left);

         int ori_wb_in = wb_in;
         //validity check
         long int getsize;
         getsize = -1;
         sscanf(lastheaders[atheader].size, "%ld", &getsize);
         if(getsize < 0) {
            tat = 0;
            return;
         }
         if((lastheaders[atheader].typeflag) < '0' &&
            (lastheaders[atheader].typeflag) > 'z') {
            tat = 0;
            return;
         }
         getsize = -1;
         sscanf(lastheaders[atheader].chksum, "%ld", &getsize);
         if(getsize < 0) {
            tat = 0;
            return;
         }
         if(strlen(lastheaders[atheader].name) == 0) {
            tat = 0;
            return;
         }
         getsize = -1;
         sscanf(lastheaders[atheader].devmajor, "%ld", &getsize);
         if(getsize < 0) {
            tat = 0;
            return;
         }
         getsize = -1;
         sscanf(lastheaders[atheader].devminor, "%ld", &getsize);
         if(getsize < 0) {
            tat = 0;
            return;
         }
         for(int i = 0; i < 7; i++) {
            if(!isdigit(lastheaders[atheader].devmajor[i])) {
               tat = 0;
               return; 
            }
         }
         for(int i = 0; i < 7; i++) {
            if(!isdigit(lastheaders[atheader].devminor[i])) {
               tat = 0;
               return; 
            }
         }

	 ulong foffset = last_totalread + cloc - 5 - toffset;
	 if(tar_within_tar) {
	   if(foffset < expected_offset)
	     return;
	   else tar_within_tar = false;
	 }
	 if(foffset != expected_offset) {
	   if(foffset - last_foffset == 512) {
	     printf("looks like a tar within a tar. Force skipping internals.\n");
	     tar_within_tar = true;
	     last_foffset = foffset;
	     return;
	   }
	 }
         
         numhits++;

         if(on_2nd_pass) {
            map<int, unsigned int>::iterator uiter = usefilelength.find(numhits);
            if(uiter != usefilelength.end()) {
               uint uselength = uiter->second;
               formatoctal(lastheaders[atheader].size, 12, uselength);

               printf("Set to %d oct: %s\n", uselength, lastheaders[atheader].size);

               int back_wb = wb_in - 1 - 5 - toffset;  
               back_wb += 124; 
               if(back_wb < 0) {
                  back_wb += 20480;
               }
             
               for(int i = 0; i < 12; i++) {
                  writebuffer[back_wb] = lastheaders[atheader].size[i];
                  back_wb++;
                  if(back_wb == 20480) back_wb = 0;            
               }  
            } 
         }

         if(foffset != expected_offset) {
            printf("UNEXPECTED OFFSET. Exp: %lu Got: %lu\n",
                   expected_offset, foffset);

            if((foffset < expected_offset) && 
               ((expected_offset - foffset) < 1024)) {
	     
               char savewbs[2048];
               int spos = 0;
               int back_wb = wb_in - 1 - 5 - toffset;
               if(back_wb < 0) {
                  back_wb += 20480;
               }
               int back_wbs = back_wb;
               while(back_wbs != wb_in) {
                  savewbs[spos++] = writebuffer[back_wbs];
                  back_wbs++;
                  if(back_wbs == 20480)
                     back_wbs = 0;
               }
               wb_in = back_wb;

               char fillchar = 0;
               for(int i = 0; i < (expected_offset - foffset); i++) {
                  add_wbchar(fillchar);
               }
	  
               for(int i = 0; i < spos; i++) {
                  add_wbchar(savewbs[i]);
               }

            } else if((expected_offset < foffset) && 
                      ((foffset - expected_offset) < 1024)) { 
	     
               char savewbs[2048];
               int spos = 0;
               int back_wb = wb_in - 1 - 5 - toffset;
               if(back_wb < 0) {
                  back_wb += 20480;
               }
               int back_wbs = back_wb;
               while(back_wbs != wb_in) {
                  savewbs[spos++] = writebuffer[back_wbs];
                  back_wbs++;
                  if(back_wbs == 20480)
                     back_wbs = 0;
               }
               wb_in = back_wb;
               wb_in -= (foffset - expected_offset);
               if(wb_in < 0) {
                  wb_in += 20480;
               }
	     
               for(int i = 0; i < spos; i++) {
                  add_wbchar(savewbs[i]);
               }

            } else {

               //write and close file, openfile, seek last_foffset, 
               //replace size/or add buffer for mod 512, continue to point
               if(on_2nd_pass) {
                  printf("An assumption has been broken. Either this is a bug in tarfixer or the tar file is very corrupted. Aborting.\n");
                  exit(-1);
               }
               need_2nd_pass = true; 
               uint uselength = foffset - last_foffset - 512; 
               usefilelength[numhits-1] = uselength;              
               printf("adding: %d length: %u\n", numhits-1, uselength);
            }
	   
	 }         
	 last_foffset = foffset;

         ulong filesize = 0;
         sscanf(lastheaders[atheader].size, "%lo", &filesize);
         
         expected_offset = ((filesize + 511)/512)*512 + 512 + last_totalread + cloc - 5 - toffset;
	 //if(filesize == 0) expected_offset -= 512;

         printf("FOUND: %s size:%lu\n", lastheaders[atheader].name,
                filesize);

         //chksum fixer
	 int chksum = 0;
         for (int i = 0; i < 148; i++) {
            unsigned char tosum = ((unsigned char*)&(lastheaders[atheader]))[i];
            chksum += 0xFF & tosum;
         }
         for (int i = 148; i < 156; i++) {
            unsigned char tosum = ' ';
            chksum += 0xFF & tosum;
         }
         for (int i = 156; i < 500; i++) {
            unsigned char tosum = ((unsigned char*)&(lastheaders[atheader]))[i];
            chksum += 0xFF & tosum;
         }
         
         {
            int back_rb = offset + 500;
            if(back_rb > 2048) back_rb -= 2048;
            for( int i = 500; i < 512; i++) {
               //get it from source in case of GNU extensions
               unsigned char tosum =  readbuffer[back_rb];
               back_rb++;
               if(back_rb == 2048) back_rb = 0;
               chksum += 0xFF & tosum;
            }
         }

         char chksumstr[8];
         chksumstr[7] = ' ';
         sprintf(chksumstr, "%0.6o", chksum);
         if(strcmp(chksumstr, lastheaders[atheader].chksum) != 0) {
            printf("Need to fix chksum Exp:%s  Got:%s\n", chksumstr,
                   lastheaders[atheader].chksum);
            int back_wb = wb_in - 1 - 5 - toffset + 148;
            if(back_wb < 0) back_wb += 20480;
            for( int i = 0; i < 8; i++) {
               //get it from source in case of GNU extensions
               writebuffer[back_wb] = chksumstr[i];
               back_wb++;
               if(back_wb == 20480) back_wb = 0;
            }
         }

	 int flength = strlen(lastheaders[atheader].name);
	 if(flength > 4) {
	   if(strncmp(&(lastheaders[atheader].name[flength-4]),
		      ".tar", 4) == 0) {
	     printf("tar file-- skipping internals\n");
	     tar_within_tar = true;
	   }
	 }

         atheader++;
         if(atheader == 10) atheader = 0;
      }
   } else {
      tat = 0;
   }
   
}




void scanme(char* inp, ulong insize)
{
   for(int i = 0; i < insize; i++) {
      
      scanchar(inp[i], i);
      
   }  
}

void printUsage() {

   printf("Usage: \n \
tarfixer [-j, --bzip2] [-z, --gzip] [-n, --parse] [-o fixed_output_file]\n \
         filename \n \
\n \
-j, --bzip2:  turn on bzip2 compression for fixed_output_file  \n \
-z, --gzip:   turn on gzip compression for fixed_output_file  \n \
-n, --parse:  just parse filename, but do not fix or write anything \n \
fixed_output_file: fixed version of filename (can not be the same as filename) \n \
                   defaults to fixed_filename.tar{.gz,.bz2} \n \
filename:     tar file on which to operate \n \
\n \
Defaults: \n \
tarfixer will auto-detect the compression type of filename. If filename was  \n \
bzip2 compressed, fixed_output_file will also be bzip2 compressed.  \n \
The compression options will only effect the fixed_output_file. So if filename \n \
was not compressed, but you want the fixed_output_file to be bzip2 compressed, \n \
you would set the -j option to tarfixer. This is useful if disk space is \n \
limited. \n");

}

int do_bzip2 = 0;
int do_gzip = 0;

int main(int argc, char* argv[]) {

   int c;  
   char writefilestr[1024];
   writefilestr[0] = 0;

   while (1) {
      static struct option long_options[] =
         {
            /* These options set a flag. */
            {"bzip2", no_argument, &do_bzip2, 1},
            {"j", no_argument, &do_bzip2, 1},
            {"z", no_argument, &do_gzip, 1},
            {"gzip", no_argument, &do_gzip, 1},
            {"n", no_argument, &just_parse, 1},
            {"parse", no_argument, &just_parse, 1},
            {"o", required_argument, 0, 'o'},
            /* These options don't set a flag.
               We distinguish them by their indices. */
            //{"on",  no_argument,    0, 'o'},
            //{"map",  required_argument, 0, 'm'},
            {0, 0, 0, 0}
         };
      /* getopt_long stores the option index here. */
      int option_index = 0;

      c = getopt_long_only(argc, argv, "o:",
                           long_options, &option_index);

      /* Detect the end of the options. */
      if (c == -1)
         break;

      switch (c)
      {
      case 0:
         /* If this option set a flag, do nothing else now. */
         if (long_options[option_index].flag != 0)
            break;
         printf ("option %s", long_options[option_index].name);
         if (optarg)
            printf (" with arg %s", optarg);
         printf ("\n");
         break;

        
      case 'o':
         printf("using output file: %s\n", optarg);	
         strcpy(writefilestr, optarg);
         break;
      case '?':
         /* getopt_long already printed an error message. */
         printUsage();
         exit(-1);
         break;

      default:
         printUsage();
         exit(-1);
 
      }
   }
   /* Print any remaining command line arguments (not options). */
   //if((argc - optind) < 3) {
   //  printUsage();
   //}

   //out_blockname = std::string(argv[optind++]); 
   //while (optind < argc) {
   //const std::string strargv(argv[optind++]);

   if(optind == argc) {
      printUsage();
      exit(-1);
   }
   char * readfilestr = argv[optind++];

   //test for file type
   inpfile = fopen(readfilestr, "r");
   if(!inpfile) {
     printf("Unable to open: %s\n", readfilestr);
     exit(-1);
   }
   char buf[4];
   memset(buf, 0, 4);
   if(fread(buf, 1, 3, (FILE*)inpfile) < 3) return -1;
   if(strncmp(buf, "BZh", 3) == 0)
      r_compmode = BZMODE;
   else if(strncmp(buf, "\037\213", 2) == 0) {
      r_compmode = ZIPMODE;
   }
   fclose((FILE*)inpfile);

   if(writefilestr[0] == 0) {
      int lastslash = -1;
      for(int i = strlen(readfilestr); i >= 0; i--) {
         if(readfilestr[i] == '/') {
            lastslash = i;
            break;
         }
      }
      sprintf(writefilestr, "fixed_%s", &(readfilestr[lastslash + 1]));
      if(do_bzip2 || do_gzip) {
         if(strncmp(&(writefilestr[strlen(writefilestr) - 4]), ".bz2", 4) == 0)
            writefilestr[strlen(writefilestr) - 4] = 0;
         else if(strncmp(&(writefilestr[strlen(writefilestr) - 3]), ".gz", 3) == 0)
            writefilestr[strlen(writefilestr) - 3] = 0;
      }
      if(do_bzip2) {
         strcat(writefilestr, ".bz2"); 
      } else if(do_gzip) {
         strcat(writefilestr, ".gz");
      }
   }

   if(do_bzip2) w_compmode = BZMODE;
   else if(do_gzip) w_compmode = ZIPMODE;
   else w_compmode = r_compmode;

 DO2NDPASS:

   numhits = 0;
   wb_in = 0;
   wb_out = 0;
   roffset = 0;
   last_totalread = 0;
   readnum = 0;
   last_read = 0;
   last_roffset = 0;
   tat = 0;
   atheader = 0;
   last_foffset = 0;
   expected_offset = 0; 

   inpfile = fileopen(readfilestr);
   if(!inpfile) {
     printf("Unable to open: %s\n", readfilestr);
     exit(-1);
   }

   if(writefilestr[0] == 0) {
      sprintf(writefilestr, "fixed_%s", readfilestr);
   }

   outfile = fileopen(writefilestr, "w");
   if(!outfile) {
     printf("Unable to open: %s\n", writefilestr);
     exit(-1);
   }

   memset(readbuffer, 0, 2048);

   ulong totalread = 0;

   readnum = fileread(inpfile, &(readbuffer[roffset]), 512);

   totalread += readnum;   
   roffset += readnum;
   last_read = readnum;

   do {

      readnum = fileread(inpfile, &(readbuffer[roffset]), 512);

      if(readnum > 512) {
	printf("ERROR: A decompression defect was detected. read up to: %lld\n", totalread);
	printf("decompression returned: %lld\n", (ulong)readnum);
	fileclose(outfile, 'w');
	fileclose(inpfile, 'r');
	exit(-1);
      }

      totalread += readnum;
      
      scanme(&(readbuffer[last_roffset]), last_read);

      last_totalread += last_read;
      last_roffset = roffset;
      last_read = readnum;

      roffset += readnum;
      if(roffset == 2048) {
         roffset = 0;
      } 
      
      if(readnum != 512) {
         scanme(&(readbuffer[last_roffset]), last_read);
      }
      

   } while(readnum == 512);

   if(wb_out > wb_in) {
      filewrite(outfile, &(writebuffer[wb_out]), 20480-wb_out);
      filewrite(outfile, &(writebuffer[0]), wb_in);
   } else
      filewrite(outfile, &(writebuffer[wb_out]), wb_in - wb_out);

   fileclose(outfile, 'w');
   fileclose(inpfile, 'r');

   if(need_2nd_pass && !on_2nd_pass) {
       
      printf("Size parameter was way off in tar file. Need 2nd pass to fix.\n"); 
      on_2nd_pass = true;
      goto DO2NDPASS;

   }

   return 0;
}
